/*******************************************************************************
* Copyright (c) 2022  Xilinx, Inc. All rights reserved.
* All rights reserved. This software is made available under
* the terms of the Eclipse Public License 2.0
* which accompanies this distribution, and is available at
* https://www.eclipse.org/legal/epl-2.0/
*
* Xilinx, Inc.
* Copyright 2022  Xilinx, Inc. All Rights Reserved.
*******************************************************************************/

'use strict';

import { spawn } from 'child_process';
import { join } from 'path';
import * as vscode from 'vscode';
import { CancellationToken, DebugAdapterTracker, DebugConfiguration, DebugSession, ProviderResult, WorkspaceFolder } from 'vscode';
import { DebugProtocol } from 'vscode-debugprotocol';

const TYPE = 'tcf-debug';

const REQUEST_LAUNCH = 'launch';
const REQUEST_ATTACH = 'attach';

const CONFIG_CONTEXT = 'context'; // only used during attach

const DEFAULT_REQUEST = REQUEST_LAUNCH;
const DEFAULT_NAME = 'Debug File';
const DEFAULT_HOST = '127.0.0.1';
const DEFAULT_TARGET = {
	host: DEFAULT_HOST,
	port: 1534
};

const CMD_DEBUG_FILE = 'extension.tcf-debug.debugEditorContents';

const SERVER_PATH = join(__dirname, "..", "server", "org.eclipse.tcf.dap.jar");

const DAP_CONFIGURATION_SECTION = {
	ROOT: 'TCFAdapter',
	LOG_TO_DEBUG_CONSOLE: 'logToDebugConsole'
} 
const DAP_CONFIGURATION = {
	LOG_TO_DEBUG_CONSOLE: false
} 

export function activate(context: vscode.ExtensionContext) {
	context.subscriptions.push(vscode.commands.registerCommand(CMD_DEBUG_FILE, (resource: vscode.Uri) => debugEditorContents(resource)));
	context.subscriptions.push(vscode.debug.registerDebugConfigurationProvider(TYPE, new TcfDebugConfigurationProvider()));
	context.subscriptions.push(vscode.debug.registerDebugAdapterDescriptorFactory(TYPE, new TcfDebugAdapterDescriptorFactory()));
	context.subscriptions.push(vscode.debug.registerDebugAdapterTrackerFactory(TYPE, new TcfDebugAdapterTrackerFactory()));
	context.subscriptions.push(vscode.workspace.onDidChangeConfiguration(workspaceConfigurationChanged));
	updateDAPConfiguration();
}

export function deactivate() {
	// nothing to do since we use context subscriptions
}

function workspaceConfigurationChanged(event: vscode.ConfigurationChangeEvent): void {
	if(event.affectsConfiguration(DAP_CONFIGURATION_SECTION.ROOT)) {
		updateDAPConfiguration();
	} 
}

function updateDAPConfiguration(config: vscode.WorkspaceConfiguration = vscode.workspace.getConfiguration(DAP_CONFIGURATION_SECTION.ROOT)): void {
	DAP_CONFIGURATION.LOG_TO_DEBUG_CONSOLE = config.get(DAP_CONFIGURATION_SECTION.LOG_TO_DEBUG_CONSOLE, false);
	console.log(DAP_CONFIGURATION.LOG_TO_DEBUG_CONSOLE);
} 

function debugEditorContents(resource: vscode.Uri): void {
	let targetResource = resource;
	if (!targetResource && vscode.window.activeTextEditor) {
		targetResource = vscode.window.activeTextEditor.document.uri;
	}
	if (targetResource) {
		vscode.debug.startDebugging(undefined, {
			name: DEFAULT_NAME,
			type: TYPE,
			request: DEFAULT_REQUEST,
			target: { ...DEFAULT_TARGET, file: targetResource.fsPath }
		});
	}
}

class TcfDebugConfigurationProvider implements vscode.DebugConfigurationProvider {
	resolveDebugConfiguration(_folder: WorkspaceFolder | undefined, config: DebugConfiguration, _token?: CancellationToken): ProviderResult<DebugConfiguration> {
		const editor = vscode.window.activeTextEditor;
		if (!config.request || config.request === REQUEST_LAUNCH) {
			return this.resolveLaunchConfiguration(config, editor?.document.uri.fsPath);
		}
		if (config.request === REQUEST_ATTACH) {
			return this.resolveAttachConfiguration(config, editor?.document.uri.fsPath);
		}
	}

	resolveLaunchConfiguration(config: DebugConfiguration, defaultFile: string | undefined): ProviderResult<DebugConfiguration> {
		config.name = config.name ?? DEFAULT_NAME;
		config.type = config.type ?? TYPE;
		config.request = config.request ?? REQUEST_LAUNCH;
		config.target = { ...DEFAULT_TARGET, ...config.target };
		config.target.file = config.target.file ?? defaultFile;
		return config;
	}

	resolveAttachConfiguration(config: DebugConfiguration, defaultFile: string | undefined): ProviderResult<DebugConfiguration> {
		const problem = this.evaluateConfiguration(config, CONFIG_CONTEXT);
		if (problem) {
			return problem;
		}
		config.name = config.name ?? DEFAULT_NAME;
		config.type = config.type ?? TYPE;
		config.host = config.host ?? DEFAULT_HOST;
		config.target = { ...DEFAULT_TARGET, ...config.target };
		config.target.file = config.target.file ?? defaultFile;
		return config;
	}

	evaluateConfiguration(config: object, property: string): Thenable<undefined> | undefined {
		if (config[property] === undefined) {
			return vscode.window.showInformationMessage(`Need to specify '${property}' in configuration to use TCF debug.`).then(() => undefined);
		}
	}
}

class TcfDebugAdapterTrackerFactory implements vscode.DebugAdapterTrackerFactory {
	createDebugAdapterTracker(session: DebugSession): ProviderResult<DebugAdapterTracker> {
		return new TcfDebugAdapterTracker(session);
	}
}

class TcfDebugAdapterTracker implements DebugAdapterTracker {
	constructor(protected session: DebugSession) {
	}

	onWillStartSession(): void {
		this.log(`[TCF-Adapter] [${this.session.id}] startDebugAdapter`);
	}

	onWillReceiveMessage(message: any): void {
		this.log(`[TCF-Adapter] [${this.session.id}] <=== ${JSON.stringify(message, undefined, 2)}`);
	}

	onDidSendMessage(message: any): void {
		this.log(`[TCF-Adapter] [${this.session.id}] ===> ${JSON.stringify(message, undefined, 2)}`);
		if (message.type === 'response') {
			const response = message as DebugProtocol.Response;
			if (response.command === 'launch') {							
				if ( response.body.includes("Connection refused") ) {										
					vscode.window.showErrorMessage(`Could not connect to TCF Agent at host ${this.session.configuration.target.host} and port ${this.session.configuration.target.port}. Make sure your target is running and check the target details in launch.json`).then(() => undefined);
				}
			}
		}
		if (message.type === 'request') {
			const request = message as DebugProtocol.Request;
			if (request.command === 'runInTerminal') {
				const runInTerminalRequest = request as DebugProtocol.RunInTerminalRequest;
				const args = runInTerminalRequest.arguments.args;
				if (args.length == 2) {
					if ('attachToDA' === args[0]) {
						const context = args[1]; 
						this.log(`[TCF-Adapter] [${this.session.id}] ---- Attaching new child ${this.session.configuration.target.host} ${this.session.configuration.target.port} ${context}`);
						vscode.debug.startDebugging(this.session.workspaceFolder, {
							name: this.session.name + " (Attach)",
							type: TYPE,
							request: DEFAULT_REQUEST,
							target: { 
								host: this.session.configuration.target.host,
								port: this.session.configuration.target.port,
								context: context,
							},
							__parentId: this.session.id
						}, this.session);
					} 
				} 
			} 
		} 
	}

	onWillStopSession(): void {
		this.log(`[TCF-Adapter] [${this.session.id}] debugAdapterWillStop`);
	}

	onError(error: Error): void {
		this.log(`[TCF-Adapter] [${this.session.id}] debugAdapterError: ${error}`);
	}

	onExit(code: number | undefined, _signal: string | undefined): void {
		this.log(`[TCF-Adapter] [${this.session.id}] debugAdapterExit: ${code}`);
	}

	protected log(log: string): void {
		console.log(log);
		if(DAP_CONFIGURATION.LOG_TO_DEBUG_CONSOLE)   {
			vscode.debug.activeDebugConsole.appendLine(log);
		} 
	} 
}

class TcfDebugAdapterDescriptorFactory implements vscode.DebugAdapterDescriptorFactory {
	createDebugAdapterDescriptor(session: vscode.DebugSession, _executable: vscode.DebugAdapterExecutable | undefined): ProviderResult<vscode.DebugAdapterDescriptor> {
		if (session.configuration.request === REQUEST_LAUNCH) {
			return launch(session);
		}
		if (session.configuration.request === REQUEST_ATTACH) {
			return attach(session);
		}
	}
}

function attach(session: vscode.DebugSession): ProviderResult<vscode.DebugAdapterDescriptor> {
	let host = session.configuration.host || DEFAULT_HOST;
	let port = session.configuration.port;
	return new vscode.DebugAdapterServer(port, host);
}

function launch(session: vscode.DebugSession): ProviderResult<vscode.DebugAdapterDescriptor> {
	return new Promise((resolve, reject) => {
	    // const serverProcess = spawn('java', ['-ea','-agentlib:jdwp=transport=dt_socket,server=y,suspend=y,address=5005','-jar', SERVER_PATH]);
		const serverProcess = spawn('java', ['-jar', SERVER_PATH]);
		// const serverProcess = spawn('java', ['-Djava.util.logging.config.file="logging.properties" ','-agentlib:jdwp=transport=dt_socket,server=y,suspend=y,address=5005','-jar', SERVER_PATH]);

		let connected = false;
		serverProcess.stderr!.on('data', (buffer: Buffer) => {
			let text = buffer.toString();
			if (!connected) {
				// parse connection info from application error output
				if (text.match(/TCF Debug Adapter/)) {
					let match = text.match(/HOST=([^\s]*)[\s]+PORT=([0-9]*)/);
					if (match) {
						connected = true;
						resolve(new vscode.DebugAdapterServer(parseInt(match[2]), match[1]));
					} else {
						reject(new Error("Unable to determine debugger host and port"));
					}
				}
			}
			vscode.debug.activeDebugConsole.append(`[TCF-Server] [${session.id}] ${text}`);
		});
		serverProcess.stdout!.on('data', (buffer: Buffer) => {
			let text = buffer.toString();
			vscode.debug.activeDebugConsole.append(`[TCF-Server] [${session.id}] ${text}`);
		});
		serverProcess.on('exit', (code) => {
			if (!connected) {
				reject(new Error(`Debugger exited without connecting (exit code ${code})`));
			}
		});
	});
}
